/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */


/*!
 *  \file main.c
 *  \brief oWL Pico reference design software.
 *
 *  Copyright (C) 2012 HD Wireless AB
 *
 *  You should have received a copy of the license along with this library.
 */


/*! \defgroup wlp_ref oWL Pico Reference Design
 *
 * The pclient application is built on top of the oWL pico API and uses the
 * owl toolkit, both provided by H&D Wireless.
 *
 * pclient provides a command line interface to directly invoke most of the
 * API functions in oWL Pico API, a ttcp throughput test tool and a small
 * web server.
 *
 * pclient can be built and used on a linux desktop PC together with the
 * SPB800 device. It can also be built for an ATMEL Xplain board, running on an
 * 8-bit CPU with 8Kb RAM. All the platform differences are handled in the
 * owl driver layer and in the main.c file.
 *
 * To build pclient on a linux system, make -C ports/linux should produce the
 * pclient binary in the ports/linux dir.
 *
 * To build pclient on for the ATMEL Xplain board, make -C ports/avr8 should
 * produce the pclient.hex binary in the ports/avr8 dir.
 *
 * The pclient reference design can be used as a starting point for using the
 * oWL Pico API for a custom application and/or on a custom platform.
 * 
 *  @{
 */
#define OWL_TRACE OWL_TRACE_DBG

#include <wlp_api.h>
#include <wlp_inet.h>

#include <owl/core/owl_defs.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_shell.h>
#include <owl/core/owl_timer.h>
#include <owl/drivers/owl_board.h>
#include <owl/drivers/owl_uart.h>
#include <owl/drivers/owl_rtc.h>
#include <owl/drivers/owl_led.h>
#include <owl/drivers/owl_spi.h>
#include <owl/drivers/owl_gpio.h>
#include <owl/components/spb800p.h>

#include <pico/wlp_shell.h>
#include <pico/wlp_ttcp.h>
#include <pico/wlp_ntp.h>
#include <pico/wlp_ping.h>
#include <pico/wlp_httpd.h>

#include "tempsensor.h"

struct httpd *http_handle = NULL;
void *owluart = NULL;

#define TIMER_PERIOD_MS 64

#ifdef __linux__
# define OWL_BOARD_SPB800P_UART ((void *) fd)
#endif

static void link_cb(void *ctx, int link)
{
        struct owlsh_ctx *sh = ctx;
        owlsh_printf(sh, "link_cb link:%d\n\r", link);
}


static void addr_cb(void *ctx, const struct ip_addr *ip)
{
        struct owlsh_ctx *sh = ctx;
        if (ip == NULL) {
                owlsh_printf(sh, "addr_cb ip:none\n\r");
                if (http_handle) {
                        httpd_destroy(http_handle);
                        http_handle = NULL;
                        owlsh_printf(sh, "httpd stopped\n\r");
                }

        } else {
                owlsh_printf(sh, "addr_cb ip:%s\n\r", inet_ntoa(ip));
                if ((http_handle = httpd_create(80)) != NULL)
                        owlsh_printf(sh, "httpd started\n\r");
        }
}


static void tick_cb(void)
{
        uint32_t period;
        period = owlrtc_get_period();
        owltmr_tick(period);
}


#if defined(OWL_BOARD_SPB800P_WAKEUP_PIN)
static void wakeup_cb(void *ctx, int enable) {

        owl_dbg("wakeup enabled == %d", enable);
        if (enable) {
                owlgpio_clr(OWL_BOARD_SPB800P_WAKEUP_PIN);
#if defined (OWL_BOARD_SPB800P_HOST_ATTENTION_PIN) && defined (HA_ACK)
                /* Wait until the SPB800 signals that it's ready. */
                while (owlgpio_is_high(OWL_BOARD_SPB800P_HOST_ATTENTION_PIN));
#else
                /* no host attention pin, hope that 5ms is enough */
                owlboard_mdelay(30);
#endif
        }
        else
                owlgpio_set(OWL_BOARD_SPB800P_WAKEUP_PIN);
}
#endif /* OWL_BOARD_SPB800P_WAKEUP_PIN */


#if defined(OWL_BOARD_SPB800P_SPI)
static void select_cb(void *spi, int select)
{
        if (select)
                owlspi_claim(spi, OWL_BOARD_SPB800P_SPI_CS);
        else
                owlspi_release(spi, OWL_BOARD_SPB800P_SPI_CS);
}
#elif defined(OWL_BOARD_SPB800P_UART)
static void select_cb(void *uart, int select)
{
# ifdef __avr8__
        static irqflags_t iflags;
        (void) uart;
        
        if (select)
                iflags = cpu_irq_save();
        else
                cpu_irq_restore(iflags);
# else
        return;
# endif
}
#endif /* OWL_BOARD_SPB800P_SPI */


#if defined(OWL_BOARD_SPB800P_SPI) && defined(__avr32__)
# include <gpio.h>
__attribute__((__interrupt__)) void spb800p_irq_handler(void)
{
        if (gpio_get_pin_interrupt_flag(OWL_BOARD_SPB800P_SPI_IRQ_PIN)) {
		gpio_clear_pin_interrupt_flag(OWL_BOARD_SPB800P_SPI_IRQ_PIN);
                wlp_irq();
	}
}
#endif


#if defined(OWL_BOARD_SPB800P_HOST_ATTENTION_PIN)
static int ha_pin_is_signalled(void)
{
        if (owlgpio_is_low(OWL_BOARD_SPB800P_HOST_ATTENTION_PIN))
                return 1;
        
        return 0;
}
#endif

static void fw_upgrade_status_cb(void *ctx, uint8_t success)
{
        struct owlsh_ctx *sh = ctx;

        if (success)
                owlsh_printf(sh, "firmware upgrade complete\n\r");
        else
                owlsh_printf(sh, "firmware upgrade failed\n\r");
}


int main(int argc, char *argv[])
{
        int err;
        struct owlsh_ctx *sh;
        int shflags = OWLSH_ESCAPECHARS;
        int demo_mode = 0;
        uint8_t fw_caps = 0;
        uint8_t mode;
        
#if defined(__avr8__) 
        demo_mode = 1;  /* connect automatically */
#elif defined(__avr32__)
        int irqline;
#elif defined(__linux__)
        int fd;
        char *devname;

        shflags |= OWLSH_NOECHO;
        
        if (argc > 1 && !owl_os_strcmp(argv[1], "-h")) {
                printf("%s [port]\n", argv[0]);
                return 0;
        }

        if(argc > 1)
                devname = argv[1];
        else
                devname = "/dev/ttyUSB0";


        if ((fd = open(devname, O_RDWR | O_NOCTTY)) < 0) {
                owl_err("Failed to open device '%s'", devname);
                return -1;
        }

#endif
        
        
        owlboard_init();
        owlrtc_init(TIMER_PERIOD_MS, tick_cb);

        if ((err = owltmr_init()) < 0) {
                owl_err("owltmr_init failed err:%d(%s)", err, owl_error(err));
                return -1;
        }
        
        /* We will use the owl_* portability layer for console
         * input/output and error/debug functionality. This allows us
         * to build this code on (for instance) a Linux PC system
         * during development. To build the code for the target 
         * host platform you need to 
         *  - Write a owl_env.h file for the host
         *  - Write a owl_uart.c file for the target platform, implementing
         *    the functions prototyped in owl_uart.h
         *  - Write a owl_led.c file for the target platform, implementing
         *    the functions prototyped in owl_led.h
         *  - Write a owl_rtc.c file for the target platform, implementing
         *    the functions prototyped in owl_rtc.h
         */

        /* Initialize the console UART */
        if (owluart_init(OWL_BOARD_SHELL_UART, UART_BAUDRATE_115200,
                         UART_DATABITS_8, UART_PARITY_NONE,
                         UART_STOPBITS_1, 0) < 0)
                return -1;
        
        /* Initialize the host console input/output system. We will
         * only use the console for status output in the example
         * (using owlsh_print()) so the console command handling will
         * be unused.
         */
        if ((sh = owlsh_create(NULL, owluart_write, OWL_BOARD_SHELL_UART,
                               shflags)) == NULL) {
                owl_err("owlsh_create failed");
                return -1;
        }
        
        /* Power on SPB800 */
        spb800p_resume();

#if defined(OWL_BOARD_SPB800P_UART)
        /* Initialize the host-WiFi UART */
        if ((err = owluart_init(OWL_BOARD_SPB800P_UART, UART_BAUDRATE_115200,
                                UART_DATABITS_8, UART_PARITY_NONE,
                                UART_STOPBITS_1, 0)) < 0) {
                owl_err("uart_init pico err:%d(%s)", err, owl_error(err));
                return -1;
        }

        
        /* Setup the owl pico library. We pass the functions
         * that the library will use to read from and write to
         * the WiFi module.
         */
        if ((err = wlp_init(WLP_INTERFACE_UART,
                            owluart_read, owluart_write, select_cb,
                            OWL_BOARD_SPB800P_UART)) < 0) {
                owl_err("wlp_init failed err:%d(%s)", err, owl_error(err));
                return -1;
        }

        /* keep track of the uart fd to support the set_baudrate command */
        owluart = OWL_BOARD_SPB800P_UART;

        
#elif defined(OWL_BOARD_SPB800P_SPI)

        /* we can now configure SPB800P_SPI_IRQ_PIN with irq function.
           
           From avr32 datasheet: "In every port there are four interrupt lines
           connected to the interrupt controller. Every eigth
           interrupts in the port are ored together to form an
           interrupt line." */
        irqline =
                AVR32_GPIO_IRQ_0 +
                (OWL_BOARD_SPB800P_SPI_IRQ_PIN / 32) * 4 +
                (OWL_BOARD_SPB800P_SPI_IRQ_PIN / 8) % 4;
        INTC_register_interrupt(&spb800p_irq_handler, irqline, AVR32_INTC_INT0);
        gpio_clear_pin_interrupt_flag(OWL_BOARD_SPB800P_SPI_IRQ_PIN);
        gpio_enable_pin_interrupt(OWL_BOARD_SPB800P_SPI_IRQ_PIN,
                                  GPIO_FALLING_EDGE);
        
        /* Initialize the host-WiFi SPI bus */
        if ((err = owlspi_init(OWL_BOARD_SPB800P_SPI,
                               OWL_SPI_MODE_MASTER)) < 0) {
                owl_err("spi err:%d(%s)", err, owl_error(err));
                return -1;
        }

        if ((err = owlspi_setup_cs(OWL_BOARD_SPB800P_SPI,
                                   OWL_BOARD_SPB800P_SPI_CS,
                                   OWL_BOARD_SPB800P_SPI_FREQ,
                                   OWL_SPI_TRANSFER_MODE_3, 16)) < 0) {
                owl_err("spi err:%d(%s)", err, owl_error(err));
                return -1;
        }

        /* Setup the owl pico library. We pass the functions
         * that the library will use to read from and write to
         * the WiFi module.
         */
        if ((err = wlp_init(WLP_INTERFACE_SPI, owlspi_read, owlspi_write,
                            select_cb, OWL_BOARD_SPB800P_SPI) < 0)) {
                owl_err("wlp_init failed err:%d(%s)", err, owl_error(err));
                return -1;
        }

#endif

        tempsensor_init();
        
#if defined(OWL_BOARD_SPB800P_WAKEUP_PIN)
        /* Register a wakeup pin control callback.
         * This is used to wake the SPB800 when Power save level 1 is
         * used. */
        wlp_set_wakeup_cb(wakeup_cb, sh);
#endif /* defined(OWL_BOARD_SPB800P_WAKEUP_PIN) */

        /* Get the firmware capabilities, so we know if the host attention pin
           is supported by the device or not */
        wlp_get_fw_caps(&fw_caps);

        /* Get the current operating mode */
        wlp_get_mode(&mode);

        /* We register callbacks so that we get notified when the
         * WiFi link is established (link up) or torn down (link down)
         * and when and IP address has been assigned to the interface
         * (since this can happen at some point after the WiFi link
         * comes up if DHCP is used).
         */
        wlp_set_link_cb(link_cb, sh);
        wlp_set_ipaddr_cb(addr_cb, sh);

        /* We register a callback to notify us if a firmware upgrade
         * completes or fails. */
        wlp_set_fw_upgrade_cb(fw_upgrade_status_cb,
                              sh);
        
        /* Add commands which can be used to control the WiFi link.
         * link_cb() will be called when the connection is up.
         */
        owlsh_addcmd(sh, &cmd_linkup, NULL);
        owlsh_addcmd(sh, &cmd_linkdown, NULL);
        owlsh_addcmd(sh, &cmd_get_hwaddr, NULL);
        owlsh_addcmd(sh, &cmd_get_network, NULL);

        /* Add commands which can be used to control the ip address settings */
        owlsh_addcmd(sh, &cmd_set_ipaddr, NULL);
        owlsh_addcmd(sh, &cmd_get_ipaddr, NULL);
        owlsh_addcmd(sh, &cmd_set_dhcp, NULL);
        owlsh_addcmd(sh, &cmd_get_dhcp, NULL);
        owlsh_addcmd(sh, &cmd_set_dhcpd, NULL);
        owlsh_addcmd(sh, &cmd_get_hostbyname, NULL);

        /* Add commands which can be used to play around with the socket
         * interface of owl pico
         */
        owlsh_addcmd(sh, &cmd_socket, NULL);
        owlsh_addcmd(sh, &cmd_bind, NULL);
        owlsh_addcmd(sh, &cmd_listen, NULL);
        owlsh_addcmd(sh, &cmd_connect, NULL);
        owlsh_addcmd(sh, &cmd_close, NULL);
        owlsh_addcmd(sh, &cmd_send, NULL);
        owlsh_addcmd(sh, &cmd_sendto, NULL);
        owlsh_addcmd(sh, &cmd_get_peeraddr, NULL);

        /* Add misc commands */
        owlsh_addcmd(sh, &cmd_get_version, NULL);
        owlsh_addcmd(sh, &cmd_get_fw_version, NULL);
        owlsh_addcmd(sh, &cmd_get_fw_caps, NULL);
        owlsh_addcmd(sh, &cmd_set_mode, NULL);
        owlsh_addcmd(sh, &cmd_get_mode, NULL);
        owlsh_addcmd(sh, &cmd_set_ps, NULL);
        owlsh_addcmd(sh, &cmd_conf_ps, NULL);
        owlsh_addcmd(sh, &cmd_set_channel, NULL);
        owlsh_addcmd(sh, &cmd_reset, NULL);
        owlsh_addcmd(sh, &cmd_set_baudrate, NULL);
        owlsh_addcmd(sh, &cmd_fw_upgrade, NULL);
        owlsh_addcmd(sh, &cmd_set_led, NULL);
        owlsh_addcmd(sh, &cmd_set_heartbeat, NULL);
        owlsh_addcmd(sh, &cmd_set_poll_period, NULL);
        
        /* Add protocol implementation examples */
        owlsh_addcmd(sh, &cmd_ttcp, NULL);
        owlsh_addcmd(sh, &cmd_ntp, NULL);
        owlsh_addcmd(sh, &cmd_ping, NULL);
        owlsh_addcmd(sh, &cmd_help, NULL);
        
#ifdef __linux__
        owlsh_addcmd(sh, &cmd_send_file, NULL);
        owlsh_addcmd(sh, &cmd_recv_file, NULL);
        owlsh_addcmd(sh, &cmd_recv, NULL);
#endif /* __linux__ */

        /* We use a default configuration on xplain for demo purposes */
        if (demo_mode && mode == WL_MODE_STA) {
                struct ip_addr ip, netmask;
                struct wl_ssid_t ssid; /* default ssid */

                IP4_ADDR(&ip, 192, 168, 2, 22);
                IP4_ADDR(&netmask, 255, 255, 255, 0);

                owl_os_strcpy(ssid.ssid, "hdwireless");
                ssid.len = owl_os_strlen("hdwireless");

                /* connect to default AP */
                wlp_linkup(&ssid, "hdwireless123", 0);

                /* set default IP address */
                wlp_set_ipaddr(&ip, &netmask, NULL, NULL);
        }
        
        /* Start the main poll loop to check the console input and making sure
         * that the owl pico API makes forward progress.
         */
        for (;;) {
                if (owluart_poll(OWL_BOARD_SHELL_UART))
                        owlsh_key(sh, owluart_inb(OWL_BOARD_SHELL_UART, 0));

                /* If the host has support for the host attention pin, we will
                   use it if also supported by the device */
#if defined(OWL_BOARD_SPB800P_HOST_ATTENTION_PIN)
                if (fw_caps & WLP_CAP_HA_PIN) {
                        /* If the host attention pin is present we
                         * only poll if the pin is asserted (active low).
                         * When the pin is asserted there's an event
                         * pending from the SPB800 module. */
                        if (ha_pin_is_signalled()) {
                                owl_vdbg("poll");
                                wlp_poll();
                        }
                        
                } else {
                        /* Without the host attention pin we need to
                         * poll the SPB800 periodically. This method will
                         * prevent the SPB800 from using power save
                         * level 1. */ 
                        wlp_poll();
                }
#else 
                wlp_poll();
#endif
                
                owltmr_poll();
                
        }                
        
        return 0;
}

/*!  @} */
