/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_ERR

#include <owl/core/owl_timer.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_err.h>

#include <wl_os.h>

struct owltmr_tmo {
        struct owltmr_tmo *next;
        owltmr_tmo_cb_t *cb;
        void *ctx;
#define TMO_EXPIRED  (1 << 0) /* fired but not deleted */
#define TMO_PERIODIC (1 << 1) /* reschedules automatically after fire */
#define TMO_DELETED  (1 << 2) /* pending delete */
        unsigned int flags;
        unsigned int ts;        /* compare timestamp */
        unsigned int intvl;     /* interval in ms */
};

struct owltmr_ctx {
        struct owltmr_tmo *head;
        volatile unsigned int cur_ts; /* current timestamp in ms */
#define OWLTMR_FLAG_POLL (1 << 0)
        int flags;
};


static struct owltmr_ctx *priv = NULL;

int
owltmr_init(void)
{
        priv = owl_os_zalloc(sizeof(struct owltmr_ctx));
        if (priv == NULL) {
                owl_err("could not alloc priv");
                return OWL_ERR_MEM;
        }
        return 0;
}

struct owltmr_tmo *
owltmr_addtmo(int ms, int periodic, owltmr_tmo_cb_t *cb, void *ctx)
{
        struct owltmr_tmo *tmo;

        owl_info("entry ms:%d periodic:%d cb:%p ctx:%p",
                 ms, periodic, cb, ctx);

        if (cb == NULL) {
                owl_err("cb is null");
                return NULL;
        }
        
        tmo = owl_os_zalloc(sizeof(struct owltmr_tmo));
        if (tmo == NULL) {
                owl_err("could not alloc tmo");
                return NULL;
        }

        tmo->cb = cb;
        tmo->ctx = ctx;
        if (periodic)
                tmo->flags |= TMO_PERIODIC;

        if (priv->head == NULL) {
                owl_dbg("set head to %p", tmo);
                priv->head = tmo;
                
        } else {
                struct owltmr_tmo *iter;
                iter = priv->head;
                while (iter->next)
                        iter = iter->next;

                owl_dbg("adding timer %p to tail of %p", tmo, iter);
                iter->next = tmo;
        }

        owltmr_modtmo(tmo, ms);
        return tmo;
}

int
owltmr_deltmo(struct owltmr_tmo *tmo)
{
        struct owltmr_tmo *iter;
        struct owltmr_tmo *prev = NULL;
        
        if (tmo == NULL) {
                owl_info("tmo is null");
                return 0;
        }

        owl_info("entry tmo:%p %s%s%s", tmo,
                 (tmo->flags & TMO_EXPIRED) ? "EXPIRED " : "",
                 (tmo->flags & TMO_PERIODIC) ? "PERIODIC " : "",
                 (tmo->flags & TMO_DELETED) ? "DELETED " : "");
        
        owl_assert(!(tmo->flags & TMO_DELETED));
                
        /* Timer is currently executing, will be removed when iter->cb()
         * returns in owltmr_poll(), now just mark as deleted
         */
        if (priv->flags & OWLTMR_FLAG_POLL) {
                tmo->flags |= TMO_DELETED;
                return 0;
        }
        
        for (iter = priv->head; iter; prev = iter, iter = iter->next) {
                if (iter != tmo)
                        continue;

                if (prev == NULL)
                        priv->head = priv->head->next;
                else
                        prev->next = tmo->next;

                owl_info("free tmo %p", tmo);
                owl_os_free(tmo);
                return 0;
        }

        owl_err("tmo not found");
        return OWL_ERR_PARAM;
}

void
owltmr_modtmo(struct owltmr_tmo *tmo, int ms)
{
        tmo->ts = priv->cur_ts;
        tmo->intvl = ms;
        tmo->flags &= ~TMO_EXPIRED;
}

unsigned int
owltmr_get_ms(void)
{
        return priv->cur_ts;
}

void
owltmr_tick(unsigned int inc)
{
        if (priv == NULL)
                return;
        priv->cur_ts += inc;
}

void
owltmr_mdelay(int ms)
{
        unsigned int old = priv->cur_ts;
        owl_dbg("busy waiting until:%d now:%d [ms]",
                old + ms, priv->cur_ts);
        while (!owltmr_intvl_passed(old, priv->cur_ts, ms));
}

void
owltmr_poll(void)
{
        struct owltmr_tmo *prev = NULL;
        struct owltmr_tmo *iter;
        
        priv->flags |= OWLTMR_FLAG_POLL; /* used in owltmr_deltmo() */
        
        for (iter = priv->head; iter; iter = iter->next) {
                if (iter->flags & TMO_EXPIRED)
                        continue;
                if (iter->flags & TMO_DELETED) {
                        continue;
                }
                if (!owltmr_intvl_passed(iter->ts, priv->cur_ts, iter->intvl))
                        continue;
                
                owl_vdbg("invoking tmo:%p", iter);
                /* set flag before invoking cb() since owltmr_modtmo() might
                 * be invoked from the cb
                 */
                iter->flags |= TMO_EXPIRED; 
                iter->cb(iter, iter->ctx);

                if (iter->flags & TMO_PERIODIC)
                        owltmr_modtmo(iter, iter->intvl);
        }

        priv->flags &= ~OWLTMR_FLAG_POLL;

        iter = priv->head;
        while (iter) {
                struct owltmr_tmo *tmo = iter;
                if (!(iter->flags & TMO_DELETED))
                        goto loop;

                if (prev == NULL)
                        priv->head = priv->head->next;
                else
                        prev->next = tmo->next;

                /* free this timer and update iter, keep prev as it
                 * is since it is still pointing to the previous
                 * valid tmo
                 */
                iter = iter->next;
                owl_info("free tmo %p", tmo);
                owl_os_free(tmo);
                continue;
        loop:
                prev = iter;
                iter = iter->next;
        }
}

int
owltmr_intvl_passed(unsigned int old, unsigned int new, unsigned int diff)
{
        if (diff && old == new)
                return 0;
        else {
                /* no wrap */
                /* timeout's scheduled with ms=0 should be invoked
                   every time owltmr_poll() is called, therefore we
                   check using >= in the comparisons below */
                if (new >= old && new - old >= diff)
                        return 1;
                /* wrap */
                else if (new < old && ((((unsigned int) -1) - old) + new) >= diff)
                        return 1;
        }

        return 0;
}

unsigned int
owltmr_get_intvl(unsigned int old, unsigned int new)
{
        /* wrap */
        if (new < old)
                return ((unsigned int) (-1) - old) + new;
        
        /* no wrap */
        return new - old;
}
