/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#ifndef OWL_TIMER_H
#define OWL_TIMER_H

/**
 * @file owl_timer.h
 * @brief Generic timer implementation.
 * 
 * This module handles timeouts and delays in a platform-independent
 * way. It executes timer-based callback functions with "minimum-time-passed"
 * semantics : Any callback is guaranteed to wait for at least the specified
 * amount of time before being called. 
 *
 * In other words, a callback may be called later than requested, but
 * never earlier.
 *
 * The timers are driven by calls to \a owltmr_poll(). This function
 * will execute any pending callbacks whose timers has expired.
 *
 * Time-measurement is based on the information passed in calls to
 * \a owltmr_tick().
 *
 * The precision of the timers are based on the frequency of the calls
 * to \a owltmr_poll() and the frequency of the calls to \a owltmr_tick()
 * and the accuracy of the timestamp passed to it.
 *
 */

struct owltmr_tmo;

typedef void (owltmr_tmo_cb_t)(struct owltmr_tmo *, void *);

/**
 * @brief Initialize the timer module.
 *
 * @return
 *   - 0 on success
 *   - OWL_ERR_MEM
 */
int owltmr_init(void);

/**
 * @brief Add timer callback.
 * 
 * The callback will run in the execution context that calls \a owltmr_poll().
 * 
 * @param ms Timeout in milliseconds.
 * @param periodic 1 if the timer is periodic. 0 if it should only execute
 *        once.
 * @param cb Callback to execute when the timer expires.
 * @param ctx Opaque context that will be passed to the callback.
 * @return 
 *   - A pointer to a timer handle on success
 *   - NULL on failure
 */
struct owltmr_tmo *owltmr_addtmo(int ms, 
                                 int periodic,
                                 owltmr_tmo_cb_t *cb, 
                                 void *ctx);

/**
 * @brief Delete a timer.
 *
 * @param tmo A timer handle to delete.
 * @return
 *   - 0 on success
 *   - OWL_ERR_PARAM if the timer was not found
 */
int owltmr_deltmo(struct owltmr_tmo *tmo);

/**
 * @brief Modify a timer.
 *
 * @param tmo A timer handle to modify.
 * @param ms New timeout in milliseconds.
 */
void owltmr_modtmo(struct owltmr_tmo *tmo, int ms);

/**
 * @brief Get the current timestamp in milliseconds.
 *
 * @return
 *    - The current time in milliseconds.
 */
unsigned int owltmr_get_ms(void);

/**
 * @brief Update the current time value.
 *
 * Tell the timer implementation what the current time is. 
 *
 * This function can be called in interrupt context.
 *
 * @param timestamp The time passed since the last call to \a owltmr_tick()
 *                  in milliseconds.
 */
void owltmr_tick(unsigned int ms);

/**
 * @brief Execute any expired timers.
 *
 * All registered, expired timer callbacks are executed by
 * this function.
 */
void owltmr_poll(void);

/**
 * Check if \a diff ms has passed between \a old and \a new. Handles
 * the case that \a new has wrapped and is now smaller than \a old.
 */
int owltmr_intvl_passed(unsigned int old, unsigned int new, unsigned int diff);

/**
 * Get the number of units passed between \a old and \a new. Handles
 * the case that \a new has wrapped and is now smaller than \a old.
 * The return value is still a postive number in this case.
 */
unsigned int owltmr_get_intvl(unsigned int old, unsigned int new);

/**
 * @brief Delay execution.
 *
 * Wait for the specified amount of time.
 *
 * This function cannot be used if owltmr_poll() is called in the
 * same executing context. Unless the millisecond time is updated
 * in another context (such as in an interrupt service routine) this
 * call will hang.
 * @param ms Time to delay in milliseconds.
 * 
 */
void owltmr_mdelay(int ms);

#endif
