/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */


/**
 * @file wlp_api.h
 *  
 * @author H&D Wireless AB
 * @brief oWL Pico API
 * 
 * \section intro Introduction
 * This is the documentation for the oWL Pico API.
 *
 * \mainpage oWL Pico API Reference Manual
 * See pico.hd-wireless.se for the lastest documentation, software and
 * tutorials.
 * 
 * oWL Pico API is a programming interface for Wi-Fi (802.11) and network
 * applications. It aims to be a complete interface for embedded
 * applications wanting to use wireless as a communications interface.
 * 
 * The oWL Pico software API has a footprint of a few kilobytes and
 * provides functions to support full Wi-Fi link management, network
 * address management and TCP, UDP and RAW sockets - resembling a
 * standard BSD socket interface.
 *
 * The oWL Pico API has been designed to meet the following goals :
 * \li Simple : The API is as simple as is practicable
 * to make it easy to use. 
 * \li Minimal size : The API is usable on very resource constrained
 * platforms. 
 * \li Portable : The API is deployable on any platform with a standards
 * compliant C compiler and a UART port.
 * \li OS independent : The API is deployable on systems using a real time
 * operating system as well as with applications running on the
 * "bare metal" of a hardware platform (that is without an operating system).
 *
 * 
 * The API is structured into these functional groups:
 *
 * \li \ref wlp_api
 * \li \ref wlp_link
 * \li \ref wlp_network
 * \li \ref wlp_socket
 * \li \ref wlp_ap
 * \li \ref wlp_ps
 *
 * \section wlp_api_princ Operation Principles
 * 
 * There are three important properties of the oWL Pico API to keep in
 * mind when programming with it.
 * 
 * The first is that oWL Pico API is \b asynchronous. For instance, when the
 * \a wlp_linkup() function is called to attempt connection with an access
 * point it will trigger a sequence of packets being exchanged with the
 * access point after which, if everything is okay, a connection has been
 * established. The \a wlp_linkup() call is asynchronous (or non-blocking)
 * which means that you don't know if the connection attempt has
 * succeeded after the call returns. You only know if the sequence was
 * successfully started or not. To find out if, and when, the connection
 * attempt was successful you must register an event handler using the
 * function \a wlp_set_link_cb(). This is true of a number of API calls
 * (which is indicated in their documentation).
 *
 * The second important property is that the oWL Pico API is \b polled. oWL
 * Pico API never executes "by itself", since it would then have to
 * support interrupts, timers, locks and other operating system dependent
 * features. Instead all asynchronous processes proceed when oWL Pico API
 * is polled by calling the \a wlp_poll() function. When \a wlp_poll() is
 * called the oWL Pico API reacts to any received packets, expires any
 * internal timers and performs any other tasks necessary for forward
 * progress. After \a wlp_poll() returns nothing will happen unless it or
 * some other oWL Pico API function is called again.
 *
 * The third important property is that oWL Pico API is \b not \b thread
 * \b safe. All oWL Pico API calls must execute in the same context since
 * the library has no knowledge of the locking mechanisms available (if
 * any).
 * 
 * \section contactinfo Contact Information
 * For further information, visit
 * <a href="http://www.hd-wireless.se/">H&D Wireless</A>.\n
 * Support and documentation: http://pico.hd-wireless.se/
 * 
 */

#ifndef WLP_API_H
#define WLP_API_H

#include <stdint.h>

#include "wlp_proto.h"


#define WLP_FAILURE      -1  /*!< General failure */
#define WLP_ERR_PARAM    -2  /*!< Parameter error */
#define WLP_ERR_MEM      -3  /*!< Out of memory */
#define WLP_ERR_CONN     -4  /*!< No connection */
#define WLP_ERR_STATE    -5  /*!< Invalid state */
#define WLP_ERR_PARSE    -6  /*!< Input parsing error */
#define WLP_ERR_PATH     -7  /*!< Path not found */
#define WLP_ERR_TYPE     -8  /*!< Incompatible type */
#define WLP_ERR_PERM     -9  /*!< No permission */
#define WLP_ERR_NOTSUPP  -10 /*!< Operation not supported */
#define WLP_ERR_RANGE    -11 /*!< Out of range */
#define WLP_ERR_SIZE     -12 /*!< Bad size */
#define WLP_ERR_AGAIN    -13 /*!< Retry later */
#define WLP_ERR_BUSY     -14 /*!< Busy */
#define WLP_ERR_DISABLED -15 /*!< Function is disabled */
#define WLP_ERR_PROTO    -16 /*!< Protocol error */
#define WLP_ERR_TIMEOUT  -17 /*!< Timeout */
#define WLP_ERR_IGNORED  -18 /*!< Ignored, no effect */
#define WLP_ERR_CANCEL   -19 /*!< Canceled by remote */
#define WLP_ERR_SYNC     -20 /*!< Sync error */
#define WLP_ERR_ABORT    -21 /*!< Aborted */
#define WLP_ERR_INUSE    -22 /*!< Already in use */
#define WLP_ERR_USAGE    -23 /*!< General usage error */
#define WLP_ERR_CHECKSUM -24 /*!< Checksum failure */

/*! \ingroup wlp_api */
#define WLP_INVALID_SOCKET_ID 0

/*! \ingroup wlp_api */
#define WLP_INTERFACE_UART 0
#define WLP_INTERFACE_SPI  1

/*! \ingroup wlp_api
 * Wireless network information structure.
 *
 *
 */
struct wlp_network_t {
        struct wl_ssid_t ssid;          /*!< The SSID of the network */
        struct wl_mac_addr_t bssid;     /*!< The BSSID of the network */
        int16_t rssi;                   /**< Received Signal Strength in dBm
                                           (measured on beacons) */
        int16_t snr;                    /**< Received Signal to noise ratio in
                                           dBm (measured on beacons) */

        enum wlp_auth_mode auth;
        uint8_t channel;
        enum wl_conn_type_t type;
};



struct wlp_network_list {
        int cnt;
        struct wlp_network_t *net;
};

/*! \defgroup wlp_api Library support functions
 *
 * These functions manage the library in general.
 *
 * @{
 *
 */ 

 
/**
 * \brief Initialize the wifi device.
 *
 * Initialize the wlp api and configure the uart settings on the wifi device.
 * Must be called before any other function in the wlp api.
 *
 * wlp_init() will send a 'w' character, a 'w' will be received once the wifi
 * device is ready. The 'w' chars must be transmitted and received at
 * 115200 8N1.
 *
 * Note again that the uart which the callback parameters readf and writef
 * uses must be configured for 115200 8N1 before calling wlp_init().
 *
 * Allow the wifi device up to 200 ms to boot before calling wlp_init() after
 * power up.
 *
 * The sequence on the host becomes
 *  1) Power on host and wifi device
 *  2) Wait 200 ms
 *  3) Call wlp_init() with host uart configured as 115200 8N1
 *  4) Done, now the wifi device is ready to use.
 *
 * @param interface selects the host communication interface. This can be either
 *              WLP_INTERFACE_UART or WLP_INTERFACE_SPI.
 * @param readf is a pointer to a function which reads chars from the bus which
 *              the wifi device is connected to and stores them in the location
 *              pointed to by the data paramter. Note that readf should block
 *              until the number of bytes specified in the len parameter are
 *              read. This function must be able to handle NULL buffer
 *              arguments, in which case the read data should be discarded.
 * @param writef should read chars from the location pointed to by the data
 *              paramter send them to the bus which the wifi device is connected
 *              to. Note that writef should block until the number of bytes
 *              specified in the len parameter are written.
 * @param selectf will be invoked before and after any write-read transaction.
 *              For UART interfaces, this function could e.g. disable/enable
 *              interrupts on the host to ensure the UART data flow. For SPI
 *              interfaces, this could assert/desassert the chip select signal
 *              connected to the device. If the select parameter is 1, the
 *              transaction will start when this function has returned. If the
 *              select parameter is 0, the transaction is completed.
 * @param ctx is an opaque context pointer that will be passed back to readf,
 *              writef and selectf.
 * 
 * @return
 *   - 0 on success
 *
 */
int wlp_init(int interface,
             int (*readf)(void *ctx, void *data, int len),
             int (*writef)(void *ctx, const void *data, int len),
             void (*selectf)(void *ctx, int select),
             void *ctx);

/**
 * \brief Register a callback that will be invoked when wlp_api is idle.
 *
 * The yield callback can be used by the host to perform other tasks while
 * waiting for a response from the device. The yield callback will be invoked
 * when any function in the wlp_api is called that involves communication with
 * the device. The yield callback can also be used to e.g. put the host in
 * sleep mode until the response from the device is ready.
 *
 * The yield callback will only be used in case the WLP_INTERFACE_SPI was used
 * in the call to wlp_init().
 *
 * @param yieldf is a pointer to a function which will be invoked when any
 *              function in the wlp_api is waiting for a response from the
 *              device.
 * @param ctx is an opaque context pointer that will be passed back to yieldf.
 *
 */
void wlp_set_yield_cb(void (*yieldf)(void *ctx), void *ctx);


/**
 * \brief Handler for interrupts from device.
 *
 * This function should be called when an interrupt is received from the device.
 *
 * Since the wlp_api functions generally commuicates with the device and
 * blocks until a reponse is received, the wlp_irq() naturally needs to be
 * called from a different context (e.g. another thread or an interrupt
 * handler) to avoid infinite blocking. However, any calls to wlp_poll() should
 * not be invoked from a different context unless concurrency protection is
 * ensured by e.g. semaphores.
 *
 */
void wlp_irq(void);


/**
 * \brief Change the baudrate of the device.
 *
 * Send a uart configuration message which will configure the wifi device
 * according to the provided baudrate and rtscts parameters.
 *
 * Upon successful return, the wifi device will use the new uart configuration.
 * The host must change its uart settings accordingly before any other function
 * in wlp_api is called.
 *
 * The sequence on the host becomes
 *  1) Call wlp_set_baudrate() to reconfigure the device baudrate.
 *  2) Wait 250 ms after wlp_set_baudrate() has returned to allow the device to
 *     reconfigure its uart.
 *  3) Reconfigure host uart baudrate.
 *  4) Done, now the wifi device is ready to use with the new baudrate setting.
 *
 * The device supports the following baudrates:
 *
 * \verbatim
 *
 *    Baudrate | Error (%)
 *    --------------------
 *       9600  | 0
 *      19200  | 0
 *      38400  | 0
 *      57600  | 0.08
 *     115200  | 0.16
 *     230400  | 0.16
 *     460800  | 0.16
 *     921600  | 0.16
 *    1000000  | 0
 *    1500000  | 0
 *    2000000  | 0
 *    3000000  | 0
 *    4000000  | 0
 *
 * \endverbatim
 *
 * Note that the actual baudrate used might differ from the desired baudrate.
 * This is due to the fractional baud rate generation. It is recommended to
 * keep the sum of the transmitter and receiver error less than 2.00 %. As shown
 * by the table above, the SPB800 has a very small fractional error on the
 * baudrates.
 * 
 * @param baudrate is the baudrate which the wifi device should switch to.
 * @param rtcts is the hardware handshake configuration which the wifi device
 *              should switch to. Set to 1 to enabled hardware handshaking, and
 *              0 to disable hardware handshaking. Any hardware handshake will
 *              use the RTS and CTS pins.
 * @param readf is a pointer to a function which reads chars from the bus which
 *              the wifi device is connected to and stores them in the location
 *              pointed to by the data paramter. Note that readf should block
 *              until the number of bytes specified in the len parameter are
 *              read.
 * @param writef should read chars from the location pointed to by the data
 *              paramter send them to the bus which the wifi device is connected
 *              to. Note that writef should block until the number of bytes
 *              specified in the len parameter are written.
 * @param ctx is an opaque context pointer that will be passed back to readf
 *              and writef.
 *
 * @return
 *   - 0 on success
 *   - WLP_ERR_PARAM if invalid baudrate or rtscts parameters.
 *
 */
int wlp_set_baudrate(uint32_t baudrate, uint8_t rtscts);


/**
 * Reset the device. This function is typically used to switch
 * betweeen AP and STA modes after wlp_set_mode() has been used.
 * wlp_init() does not have to be called after wlp_reset().
 *
 * NOTE: In this release the baudrate must be 115200 before this function is
 * called. See wlp_set_baudrate() for details on how to change the baudrate.
 *
 */
int wlp_reset(void);

#if WLP_STACK_OFFLOAD
/**
 * \brief Prepare the device for a firmware upgrade operation.
 * 
 * @param method Specify how to retrieve the firmware image.
 *               The only currently supported method is 
 *               WLP_UPGRADE_METHOD_HTTP_POST.
 * @param addr IP address. This parameter is ignored with method 
 *             WLP_UPGRADE_METHOD_HTTP_POST.
 * @param port Port number.
 *             With WLP_UPGRADE_METHOD_HTTP_POST the \a port parameter
 *             specifies the port on which the upgrade web page will
 *             be presented.
 * @param path Path. This parameter is ignored with method
 *             WLP_UPGRADE_METHOD_HTTP_POST.
 * @return
 *     - 0 on success
 *     - WLP_ERR_PARAM
 */
int wlp_fw_upgrade(uint8_t method, 
                   struct ip_addr *addr, 
                   uint16_t port,
                   const char *path);

/**
 * \brief Register a firmware upgrade notification callback.
 * 
 * If wlp_fw_upgrade() has been called then the callback registered
 * with this call will be invoked to signal success of failure of
 * the upgrade process.
 *
 * @param status_cb Callback to be invoked when the upgrade process has
 *                  been completed. \a success will be 1 if the upgrade
 *                  was successful or 0 if the upgrade failed.
 * @return
 *     - 0 on success
 */
int wlp_set_fw_upgrade_cb(void (*status_cb)(void *ctx, uint8_t success),
                          void *ctx);


#endif /* WLP_STACK_OFFLOAD */

/**
 * \brief Get the firmware version.
 *
 * @param code will contain the fw version code upon successful return. The
 *              version can be compared to the driver version (WLP_VERSION_CODE)
 *              using the WLP_VERSION(), WLP_MAJOR_VERSION(),
 *              WLP_MINOR_VERSION() and the WLP_PATCH_VERSION() macros.
 *
 * @param build will hold the build information upon successful return. The
 *              build information is a string of information for the particular
 *              fw build used on the device. No more than \param len characters,
 *              including null termination, will be copied to the string pointed
 *              to by build. This parameter can be NULL.
 *
 * @param build_len   is the maximum amount of chars, including null termination,
 *              that will be copied to \param build.
 *
 * @param hw_rev will hold the hardware revision string upon successful
 *              return. The hardware revision string describes the particular
 *              PCB revision and model of the pico server device. \a hw_rev
 *              has the format "<rev>:<board>", or is an empty string if
 *              the revision information is not available on the current
 *              pico device. No more than \param len characters, including
 *              null termination, will be copied to the string. This parameter
 *              can be NULL.
 *
 * @param hw_rev_len   is the maximum amount of chars, including null termination,
 *              that will be copied to \param hw_rev.
 * @return
 *   - 0 on success
 *   - WLP_FAILURE if any failure occurs
 *
 *
 */
int wlp_get_fw_version(uint32_t *code, 
                       char *build, 
                       int build_len,
                       char *hw_rev, 
                       int hw_rev_len);

/**
 * \brief Get WiFi device capabilities.
 *
 * Some functionality is dependent on the capabilities of the
 * WiFi device that wlp_api interfaces with. \a wlp_get_fw_caps()
 * retrieves a flag field describing the capabilities of the
 * attached WiFi device.
 *
 * The following capabilities are supported:
 * 
 * - WLP_CAP_HA_PIN
 *
 *   The device supports a Host Attention pin. The presence of the
 *   Host Attention pin allows the WiFi device to alert the host when
 *   there is pending activity and \a wlp_poll() should be
 *   called. This allows the host to enter power save states if it can
 *   be woken up when the host attention pin is asserted. It also
 *   allows the WiFi device to spend longer times in power save state
 *   (since calling wlp_poll() will wake it up). The host attention
 *   pin is active low so that a low voltage on the pin means that the
 *   host should call \a wlp_poll() repeatedly until the host
 *   attention pin voltage changes to high. If the WLP_CAP_HA_PIN is
 *   not present in the \a cap_flags field then the host attention pin
 *   will not be manipulated by the WiFi device and wlp_poll() should
 *   be called as often as is practicable. The host cannot sleep for
 *   long periods if this is the case. See the \ref wlp_ps section for
 *   a further discussion on how to use this flag.
 *
 * - WLP_CAP_WAKEUP_PIN
 *
 *   The device supports a Wakeup pin. The presence of the Wakeup pin
 *   allows the WiFi device to enter deeper power save modes where
 *   the Wakeup pin is used to wake it up from sleep.
 *   

 * @param cap_flags Contains a bitmask of the supported
 *        capabilities flags on output.
 * @return 
 *   -  0 on success
 */
int wlp_get_fw_caps(uint8_t *cap_flags);



/*! @} */


/*! \defgroup wlp_link Wi-Fi management functions
 *
 * These functions manage the Wi-Fi connection such as encryption
 * keys, MAC addresses, and connecting to access points.
 *
 * @{
 *
 */ 


/**
 * \brief Start a scan on all channels.
 *
 * The scan result can be retrieved by \a wlp_get_network_list()
 * once the \a complete_cb() callback has been called.
 *
 * Note : Scanning may interfere with the connection management on
 *        some WiFi modules. If the module is already scanning because
 *        it is trying to connect to a network then wlp_scan() may
 *        return WLP_ERR_BUSY. To get manual scanning to work in
 *        that situation you may have to issue a wlp_linkdown() call
 *        first to stop the connection manager from scanning.
 *
 * @param ssid SSID to scan for. Can be NULL, in which case all SSIDs
 *       are included in the result (subject to memory restrictions).
 * @param complete_cb Callback that will be invoked once the scan operation
 *       is complete.
 * @param ctx is an opaque context pointer that will be passed back to 
 *        complete_cb.
 */
int wlp_scan(struct wl_ssid_t *ssid,
             void (*complete_cb)(void *ctx, int cnt), void *ctx);

int wlp_get_network_list(struct wlp_network_list **netlist);

/**
 * \brief Set a wifi link status callback function.
 *
 * Register a callback that will be invoked when the wifi connection status
 * changes.
 *
 * Every time wlp_poll() is invoked, the callback condition will be evaluated
 * and any registered callback will be invoked from same context as the call to
 * wlp_poll().
 *
 * Note that the link callback will be invoked when the wifi device is
 * connected to the access point, an IP address does not have to be
 * configured. See wlp_set_ipaddr_cb() for information on how to get notified 
 * when an IP address is configured.
 *
 * Note that the connection status will also be indicated by any wlp_api that
 * performs socket communication (e.g. wlp_send() and wlp_recv()) through
 * their return values, event though wlp_set_link_cb() is not called.
 *
 * In access point mode, the link callback will be invoked when the first
 * client is connected to the access point. See wlp_set_mode() for more
 * information on how to set access point mode.
 *
 * @param link_cb will be invoked upon a link status change. The callback
 *              parameter \a link will be set to 1 to indicate that the link
 *              has been established and 0 to indicate that the link has been
 *              shut down.
 * @param ctx is an opaque context pointer that will be passed back to link_cb.
 *
 */
void wlp_set_link_cb(void (*link_cb)(void *ctx, int link), void *ctx);


/**
 * \brief Connect the wifi link.
 *
 * Enable the wifi link. Upon successful return, the wifi device will try
 * to establish and maintain a connection to the network specified by the
 * parameters. When the link status (connected, disconnected) changes, the
 * caller will be notified of the new status through the callback registered
 * in wlp_set_link_cb(). Information about the current network can be obtained
 * through wlp_get_network().
 *
 * Note again that successful return means that the wifi device will now
 * _initate_ a link using the desired parameters, not that the link
 * is already established.
 *
 * If wlp_linkup() is called when already connected to an access point, the
 * wifi device will try to switch to the network configuration according to
 * the provided parameters.
 *
 * wlp_linkup() will only manage the link layer, not the network or socket
 * layer. See wlp_set_dhcp() and wlp_set_ipaddr() for information about
 * configuring the network layer and wlp_socket() for information about the
 * socket layer.
 *
 * In access point mode, this function will start an access point with the
 * given ssid and key parameters. See wlp_set_mode() for more information on
 * how to set access point mode.
 *
 * If the \a auth parameter is WLP_AUTH_AUTO WPA2 authentication with CCMP encryption
 * will be used for security (in access point).
 *
 * @param ssid is the name of the access point to connect to or the name
 *              of the access point to start (in access point mode).

 * @param key is the encryption key to use. If this parameter is NULL,
 *              no enryption will be used. Should be provided as an
 *              ascii string. In case of a WPA key (parameter auth set
 *              to WLP_AUTH_AUTO, WLP_AUTH_WPA_PSK or
 *              WLP_AUTH_WPA2_PSK), the WPA passphrase should be
 *              provided, e.g. "mysecretkey".  A WPA pass-phase must
 *              be between 8 and 63 characters long.  In case of a WEP
 *              key (parameter auth set to WLP_AUTH_WEP), the key
 *              should be provided as a string containing the
 *              hexadecimal chars (10 or 26 chars, depending on
 *              whether a 40-bit or 104-bit key is used),
 *              e.g. "a1b2c3d4e5".

 * @param auth specifies the authentication mode to use. For WEP this
 *             parameter must be set to WLP_AUTH_WEP. For WPA/WPA2 the
 *             parameter can be set to WLP_AUTH_AUTO to select the
 *             mode from the network, or to WLP_AUTH_WPA_PSK or
 *             WLP_AUTH_WPA2_PSK to explicitly select WPA or
 *             WPA2/RSN. WLP_AUTH_AUTO also allows open authentication
 *             (no encryption) and for security reasons should not be
 *             used when the actual authentication mode of the network
 *             is known.
 *
 * @return
 *   - 0 on success
 *   - WLP_ERR_PARAM if invalid parameters, e.g. invalid key length.
 *   - WLP_ERR_FAILURE on any other error.
 *
 */
int wlp_linkup(struct wl_ssid_t *ssid, const char *key, enum wlp_auth_mode auth);


/**
 * \brief Disconnect the wifi link.
 * 
 * Disable the wifi link. The wifi device will no longer try to maintain the
 * network connection. Any link callback that was registered through
 * wlp_set_link_cb() will be invoked when disconnection is completed,
 * if currently connected to an access point.
 *
 * Note again that successful return means that the wifi device will now
 * _initate_ a link disconnection, not that the link is already disconnected.
 *
 * If wlp_linkup() is called when already disconnected, any ongoing connection
 * attempts will be stopped.
 *
 * wlp_linkdown() will only manage the link layer, not the network or socket
 * layer. See wlp_set_dhcp() and wlp_set_ipaddr() for information about
 * configuring the network layer and wlp_socket() for information about the
 * socket layer.
 *
 * In access point mode, this function will stop the access point and it will
 * no longer transmit beacons. All clients will be disconnected. See
 * wlp_set_mode() for more information on how to set access point mode.
 *
 * NOTE: wlp_linkdown() is not supported in this release. To stop the access
 *       point, wlp_reset() can be used as a workaround.
 *
 * @return
 *   - 0 on success
 *
 */
int wlp_linkdown(void);


/**
 * \brief Get the interface MAC address.
 *
 * Return the 802.3 MAC address of the network interface.
 * 
 * @param hwaddr will hold the wifi device MAC address upon successful return.
 * @return
 *   - 0 on success
 *   - A negative error code on failure
 *
 */
int wlp_get_hwaddr(struct wl_mac_addr_t *hwaddr);


/**
 * \brief Get the 802.11 network.
 * 
 * Get the description of the currently associated network, if any.
 *
 * If associated, the device will update the RSSI and SNR information each
 * time this function is called by listening for beacon frames from the
 * current AP.
 *
 * In access point mode, this function will return the description of the
 * device when the link is up, e.g. first client connected. See wlp_set_mode()
 * for more information on how to set access point mode. In this mode, the
 * RSSI and SNR values will be set to 0.
 *
 * 
 *
 * @param net will hold the network description upon successful return.
 * @return
 *   - 0 on success
 *   - A negative error code on failure
 *
 */
int wlp_get_network(struct wlp_network_t *net);



/*! @} */


#if WLP_STACK_OFFLOAD

/*! \defgroup wlp_network Network address functions
 *
 * These functions manage the network address configuration.
 *
 * @{
 *
 */ 

/**
 * \brief Register a callback that will be invoked when an IP address is set.
 *
 * The provided callback will be invoked when an IP address is set or removed.
 * Every time wlp_poll() is invoked, the callback condition will be evaluated
 * and any registered callback will be invoked from same context as the call to
 * wlp_poll().
 *
 * If a static IP address is configured (through wlp_set_ipaddr()), this
 * function will be called at the same time as the link callback registered in
 * wlp_set_link_cb().
 *
 * If a dynamic IP address is configured (through wlp_set_dhcp()) this function
 * will be called when the link is up and an address has been successfully
 * obatined through DHCP. It will also be called when the wifi link goes down
 * (since the IP address will be removed at that time) or if the IP
 * address is relesed through a call to set_dhcp() with argument 0.
 *
 * Note that the connection status will also be indicated by any wlp_api that
 * performs socket communication (e.g. wlp_send() and wlp_recv()) through
 * their return values, even though wlp_set_ipaddr_cb() has not been called.
 *
 * @param addr_cb is called when the IP address is set or removed. If the
 *              address is set, the current address will be provided in the addr
 *              parameter. If the address is removed, the addr parameter will
 *              be NULL.
 * @param ctx is an opaque context parameter that will be passed back to the
 *              addr_cb.
 *
 */
void wlp_set_ipaddr_cb(void (*addr_cb)(void *ctx, const struct ip_addr *addr),
                       void *ctx);


/**
 * \brief Set a static IP address
 *
 * Disable DHCP and configure the device static IP address settings.
 * wlp_set_ipaddr() can be invoked either before or after the wifi link is up.
 *
 * If a static IP address is configured using this function, any callback
 * registered with wlp_set_ipaddr_cb() will be invoked as soon as the
 * wifi link status changes (see wlp_linkup()).
 *
 * @param ip is the IP address to set, e.g. 192.168.1.100. Not used if NULL.
 * @param mask is the network mask, e.g. 255.255.255.0. Not used if NULL.
 * @param gw is the gateway to use, e.g. 192.168.1.1. Not used if NULL.
 * @param dns is the dns server to use, e.g. 192.168.1.1. Not used if NULL.
 *
 * @return
 *   - 0 on success.
 *
 */
int wlp_set_ipaddr(const struct ip_addr *ip, const struct ip_addr *netmask,
                   const struct ip_addr *gw, const struct ip_addr *dns);


/**
 * \brief Get the current IP address parameters
 *
 * The currently assigned IP address setttings for the wifi device will be
 * fetched. If DHCP is enabled (through wlp_set_dhcp()) and the IP configuration
 * has not yet been assigned, all IP address values will be IP_ADDR_ANY.
 * If DHCP is disabled the statically assigned IP addresses will be fetched.
 *
 * @param ip will hold the current IP address upon successful return. Not used
 *              if NULL.
 * @param netmask will hold the current netmask upon successful return. Not used
 *              if NULL.
 * @param gw will hold the current default gateway address upon successful
 *              return. Not used if NULL.
 * @param dns will hold the current dns server address upon successful return.
 *              return. Not used if NULL.
 *
 * @return
 *   - 0 on success
 *   - WLP_ERR_CONN if not connected (IP address configuration will still
 *              be filled in into the provided parameters).
 *
 */
int wlp_get_ipaddr(struct ip_addr *ip, struct ip_addr *netmask,
                   struct ip_addr *gw, struct ip_addr *dns);


/**
 * \brief Request or release IP address through DHCP.
 *
 * Enable or disable DHCP for IP address configuration. If this function
 * is called when the wifi link is up, an IP address will be requested or
 * released upon successful return, depending on the value of the enable
 * parameter. See wlp_set_ipaddr_cb() on how to get notified when the IP address
 * configuration is set or removed.
 *
 * wlp_set_dhcp() can be invoked either before or after the wifi link is up.
 * If this function is called when the wifi link is down, any DHCP requests
 * will be performed when the wifi link comes up.
 *
 * If DHCP is enabled through this function, any configuration set through
 * wl_set_ipaddr() will not be used.
 * 
 * Note that the current IP address settings can be obtained through
 * wl_get_ipaddr() at any time.
 *
 * @param enable determines if DHCP should be enabled or disabled.
 *
 * @return 0 on success.
 *
 */
int wlp_set_dhcp(int enable);


/**
 * \brief Get the current DHCP configuration
 *
 * Get information on whether DHCP is enabled or not. Note that wlp_get_dhcp()
 * will not indicate whether an actual IP address has been obtained through DHCP
 * or not. See wlp_set_ipaddr_cb() for information about IP address
 * notifications.
 *
 * @param enabled will hold the DHCP configuration upon successful return. Not
 *              used if NULL.
 * @return
 *   - 0 on success
 *   - WLP_ERR_CONN if not connected (DHCP configuration will still be
 *              filled in into the provided parameter).
 *
 */
int wlp_get_dhcp(int *enabled);


/**
 * \brief Start or stop the device DHCP server.
 *
 * Enable or disable DHCP server for IP address configuration.
 *
 * wlp_set_dhcpd() can be invoked either before or after the wifi link is up.
 * If this function is called when the wifi link is down, any DHCP services
 * will be performed when the wifi link comes up.
 *
 * Technically, it is possible to enable the DHCP server in STA mode or AP
 * mode. However, it seldom makes sense to use a DHCP server in STA mode.
 *
 * The DHCP server will offer IP addresses in the range x.y.z.100 to x.y.z.255.
 * For instance; if the ip address of the device is 192.168.1.1 and
 * netmask 255.255.255.0 is used, the DHCP server will offer IP addresses in
 * the range 192.168.1.100 to 192.168.1.255. If the netmask was 255.255.0.0 the
 * range would be the same. In case the netmask is something like 255.255.255.y,
 * where y != 0, we would mask out the some of the IP addresses in the above
 * range to make sure that all offered IP addresses are actually valid on the
 * given network.
 *
 * When using the DHCP server, a static IP address must be assigned to
 * the device. The gateway address that is set with \a
 * wlp_set_ipaddr() will be used as the address in the "router" DHCP
 * option. Some systems (notably IOS devices) will not bring the link
 * up completely unless this option is a valid address so it is
 * usually wise to set the gateway address to the same static IP
 * address that is used for the DHCP Server itself (unless you know
 * that it should be set to something particular).
 *
 * @param enable determines if the DHCP server should be enabled or disabled.
 *
 * @return
 *   - 0 on success.
 *
 */
int wlp_set_dhcpd(int enable);

    
/*! @} */


/*! \defgroup wlp_socket Network socket functions
 *
 * These functions manage the network sockets.
 *
 * @{
 *
 */ 

/**
 * \brief Creates an endpoint for communication and returns a socket id.
 *
 * To create a TCP socket, set the type to WLP_SOCK_STREAM.
 * To create a UDP socket, set the type to WLP_SOCK_DGRAM.
 * To create a RAW socket, set the type to WLP_SOCK_RAW.
 *
 * Sockets of type WLP_SOCK_STREAM must be in a connected state
 * before any data may be sent or received on it. A connection to
 * another socket is created with wlp_connect(). Once connected, data
 * may be transferred using wlp_recv() and wlp_send() calls. When a
 * session has been completed, wlp_close() can be called. Sockets of
 * type WLP_SOCK_STREAM always uses the TCP protocol. WLP_SOCK_STREAM type
 * sockets does not preserve message boundaries, i.e sequential calls to
 * wlp_send() might result in a single, aggregated, IP packet being transferred
 * on the wireless interface. It is also possible that a single call to
 * wlp_send() results in multiple IP packets being transferred on the wireless
 * interface.
 *
 * WLP_SOCK_DGRAM and WLP_SOCK_RAW sockets are connectionless and
 * allows the sending of datagrams to correspondents named in
 * wlp_sendto(). Further, these sockets does preserve message
 * boundaries, i.e sequential calls to wlp_send() or wlp_sendto()
 * might will result in multiple IP packets being transferred on the
 * wireless interface. See wlp_send() and wlp_sendto() for more
 * information.
 *
 * The protocol parameter specifies a particular protocol to be used
 * with the socket. For the WLP_SOCK_STREAM and WLP_SOCK_DGRAM types,
 * only a single protocol exists to support a particular socket type,
 * in which case the protocol parameter is ignored. However, for
 * WLP_SOCK_RAW many protocols exists, in which case a particular protocol
 * must be specified in this manner.
 *
 * The default Wi-Fi device configuration allows for up to eight WLP_SOCK_STREAM
 * sockets (four in listening mode and four in connection mode), four
 * WLP_SOCK_DGRAM sockets, and four WLP_SOCK_RAW sockets.
 *
 * Note that is is only possible to create a single WLP_SOCK_RAW socket for
 * each protocol type.
 *
 * @param type is the socket type; WLP_SOCK_STREAM, WLP_SOCK_DGRAM or
 *              WLP_SOCK_RAW.
 * @param proto specifies a particular protocol to be used with the
 *              socket, ignored if type is WLP_SOCK_STREAM or WLP_SOCK_DGRAM.
 *
 * @return
 *   - The id of the created socket on success (> 0)
 *   - WLP_ERR_MEM if out of memory or resources (on host or wifi device)
 *   - WLP_ERR_PARAM if invalid parameters
 *   - WLP_ERR_INUSE if trying to multiple sockets with the same protocol,
 *                only for WLP_SOCK_RAW.
 *
 */
int wlp_socket(int type, int proto);


/**
 * \brief Register a connection status callback
 *
 * wlp_set_conn_cb() will register a function that will be invoked when the
 * socket connection status for the specified socket changes.
 *
 * The connection status will change when wlp_connect() is finished processing
 * or when a the remote end aborts the connection. See wlp_connect() for
 * details.
 *
 * Every time wlp_poll() is invoked, the conn_cb callback conditions
 * will be evaluated and any registered callback will be invoked from same
 * context as the call to wlp_poll().
 *
 * A registered callback can be removed by calling wlp_set_conn_cb() with
 * NULL parameters.
 *
 * @oaram sockid is the socket for which the conn_cb should be bound to.
 * @param conn_cb is a callback function that will be invoked when the
 *              connection is established or lost. The socket id and the
 *              connection state will be passed as arguments to the callback.
 *              Not used if NULL.
 * @param ctx is an opaque context parameter that will be passed back to the
 *              conn_cb.
 *
 * @return
 *   - 0 on success
 *   - WLP_ERR_PARAM if invalid parameters
 *
 */
int wlp_set_conn_cb(int sockid,
                    void (*conn_cb)(void *ctx, int sockid, int connected),
                    void *ctx);

/**
 * \brief Register a data pending callback
 *
 * Any registerd recv_cb() will be invoked when there is pending data to read
 * from the specified socket. See wlp_recv() for more information.
 *
 * Every time wlp_poll() is invoked, the recv_cb callback conditions
 * will be evaluated and any registered callback will be invoked from same
 * context as the call to wlp_poll().
 *
 * @oaram sockid is the socket for which the recv_cb should be bound to.
 * @param recv_cb is a callback function that will be invoked when there is
 *              pending data to read from the socket. The number of bytes
 *              available for reading will be passed in the len parameter.
 *              If 0, then the connection was closed by the remote end.
 *              Not used if NULL.
 * @param ctx is an opaque context parameter that will be passed back to the
 *              recv_cb.
 *
 * @return
 *   - 0 on success
 *   - WLP_ERR_PARAM if invalid parameters
 *
 */
int wlp_set_recv_cb(int sockid,
                    void (*recv_cb)(void *ctx, int sockid, int len),
                    void *ctx);


/**
 * \brief Assign an address for a socket
 *
 * When a socket is created with wlp_socket(), it has no address assigned to it.
 * wlp_bind() assigns the address specified by addr and port to the socket
 * referred to by the sockid parameter.
 *
 * A socket must be disconnected in order to use wlp_bind() on the socket.
 *
 * For sockets of type WLP_SOCK_STREAM, the address and port set through
 * wlp_bind() will be used in wlp_listen(). Also, the source address of any
 * transmitted data will be set to the address configured through wlp_bind().
 * If wlp_bind() is not called, any transmitted data will have a random source
 * port. Incoming connections for the socket will only be allowed on the port
 * which the socket is bound to.
 *
 * For sockets of type WLP_SOCK_DGRAM, the source address of any
 * transmitted data will be set to the address configured through wlp_bind().
 * If wlp_bind() is not called, any transmitted data will have a random source
 * port. Further, incoming data will only be forwarded to this socket if the
 * destination port of the data matches the port the socket is bound to. Note
 * that if the socket was connected through wlp_connect(), incoming data will
 * be forwarded to the socket based on the source address and port of the data.
 * See wlp_connect() for more information.
 *
 * For sockets of type WLP_SOCK_RAW, the the source address of any
 * transmitted data will be set to the address configured through wlp_bind().
 * The bound address will not affect which incoming data that is forwarded to
 * the socket, that is only depending on the protocol used.
 *
 * It is allowed to invoke this function multiple times on the same socket to
 * bind the socket to a new address.
 *
 * @param sockid is the socket to bind
 * @param addr must be IP_ADDR_ANY.
 * @param port is the port to bind to.
 * @return
 *   - 0 on success
 *   - WLP_ERR_PARAM if invalid parameters, e.g sockid not found.
 *   - WLP_ERR_STATE if socket is in invalid state.
 *   - WLP_ERR_NOTSUPP if the operation is not supported on the socket.
 *   - WLP_ERR_INUSE if the port is already bound by another socket.
 */
int wlp_bind(int sockid, const struct ip_addr *addr, uint16_t port);


/**
 * \brief Configure the socket to listen for incoming connections
 *
 * wlp_listen() marks the socket referred to by sockid as a passive socket,
 * that is, as a socket that will be used to accept incoming connection requests
 * wlp_accept(). Only sockets of type WLP_SOCK_STREAM may be put in listening
 * mode.
 *
 * Upon successful return the socket will listen on the port configured in
 * wlp_bind() and the listen_cb will be invoked upon any incoming connection.
 * Every time wlp_poll() is invoked, the callback condition will be evaluated
 * and any registered listen callback will be invoked from same context as the
 * call to wlp_poll().
 *
 * Note that any incoming connections will be silently rejected before the
 * listen_cb is invoked in case the wifi device is low on resources (e.g.
 * the maximum number of connected WLP_SOCK_STREAM sockets are already in use).
 *
 * @param sockid is a socket id that refers to a socket of type WLP_SOCK_STREAM.
 * @param backlog defines the maximum length to which the queue of pending
 *              connections for the socket may grow. If a connection request
 *              arrives when the queue is full, the client may receive an
 *              connection refused indication.
 * @param listen_cb will be invoked when there's a pending connection to
 *              accept. The id of the listening socket will be passed to the
 *              listen_cb. See wlp_accept() for more information.
 * @param ctx is an opaque context pointer that will be passed back to the
 *              listen_cb.
 *
 * @return
 *   - 0 on success
 *   - WLP_ERR_PARAM if invalid parameters, e.g sockid not found.
 *   - WLP_ERR_STATE if the socket is in invalid state, e.g. not bound.
 *   - WLP_ERR_NOTSUPP if the operation is not supported on the socket
 *   - WLP_ERR_MEM if out of memory or resources on the wifi device
 *
 */
int wlp_listen(int sockid, int backlog,
               void (*listen_cb)(void *ctx, int sockid),
               void *ctx);

/**
 * \brief Initiate a connection on a socket
 *
 * If the socket is of type WLP_SOCK_DGRAM or WLP_SOCK_RAW then IP and port is
 * the address to which datagrams are sent to by default, and the only address
 * from which datagrams are received.
 *
 * If the socket is of type WLP_SOCK_STREAM, this call attempts
 * to make a connection to the socket with the given address. Any conn_cb
 * function registered through wlp_set_conn_cb() will be invoked when the
 * attempt is complete, indicating wheter the connection is established or not
 * through the connected paramater. If the connected parameter is 0 when the
 * conn_cb callback is invoked, the the wifi device will not retry the
 * connection and the socket is considered disconnected again. If the connected
 * parameter is 1, then the socket is successfully connected and data transfer
 * can start by e.g. using the wlp_send() and wlp_recv() functions.
 *
 * For sockets of type WLP_SOCK_DGRAM and WLP_SOCK_RAW, there is no actual
 * concept of a "connected" and "disconnected" sockets since those sockets are
 * connectionless. However, it is still possible to call wlp_connect() for
 * these sockets, indicating that all data sent through wlp_send() will be
 * sent to the address specified in wlp_connect(). Any conn_cb registered
 * through wlp_set_conn_cb() will be invoked upon a successful return from
 * wlp_connect(). Note that a connected socket will only accept incoming data
 * which has a source address and port which matches the parameters passed
 * to wlp_connect(). See wlp_bind() for more information regarding which packets
 * that gets formwarded to the socket.
 *
 * For sockets of type WLP_SOCK_RAW, the port parameter will be ignored since
 * it is not included in the IP header.
 *
 * On a connected socket any registerd recv_cb() will be invoked when there is
 * pending data to read from the socket. See wlp_recv() and wlp_set_recv_cb()
 * for more information.
 *
 * Note that the socket connection status will also be indicated by
 * any wlp_api calls that performs socket communication
 * (e.g. wlp_send() and wlp_recv()) through their return values, even
 * though any conn_cb was not registered through wlp_set_conn_cb().
 *
 * wlp_connect() can be called multiple times for a single socket but the
 * socket must be disconnected prior to each call.
 *
 * @param sockid is the socket id to connect
 * @param ip is the IP address of the remote end
 * @param port is the port number of the remote end, ignored for sockets of
 *         type WLP_SOCK_RAW.
 *
 * @return
 *   - 0 on success
 *   - WLP_ERR_PARAM if invalid parameters, e.g sockid not found.
 *   - WLP_ERR_CONN if no network address is set or the wifi link is down.
 *   - WLP_ERR_STATE if the socket is in invalid state, e.g. not
 *              disconnected.
 *   - WLP_ERR_NOTSUPP if the operation is not supported on the socket
 *   - WLP_ERR_MEM if out of memory or resources on the wifi device
 *
 */
int wlp_connect(int sockid, const struct ip_addr *ip, uint16_t port);

/**
 * \brief Accept a connection on a socket.
 *
 * wlp_accept() will accept any pending connection on a listening socket.
 * A new socket will be created and returned as a result of this call. The
 * original listening socket is unaffected by this call.
 *
 * This function should typically be called when the callback registered in
 * wlp_listen() for the specific socket is invoked. The listening socket id
 * (as passed to wlp_listen()) should be passed to wlp_accept().
 *
 * However it is also valid to wlp_accept() even through a listen_cb is not
 * invoked for the socket (or not registerd at all). Therefore, it is also
 * possible to accept a pending connection from a specific listening socket in
 * a polling fashion.
 *
 * Upon returning from wlp_accept() the new socket is connected and ready to
 * send and receive data using e.g wlp_send() and wlp_recv(). If the
 * connection is lost, the any conn_cb registered through wlp_set_conn_cb()
 * will be invoked. Any recv_cb registered through wlp_set_recv_cb() will be
 * invoked when there is pending data to read from the socket. See wlp_recv()
 * for more information.
 *
 * @param sockid is the listening to socket to accept the new connection from.
 * @return
 *   - Returns the new socket for the connection (> 0) on success
 *   - WLP_ERR_PARAM if invalid parameters, e.g sockid not found.
 *   - WLP_ERR_STATE if the socket is in invalid state, e.g. not
 *              accepting.
 *   - WLP_ERR_NOTSUPP if the operation is not supported on the socket
 *
 */
int wlp_accept(int sockid);


/**
 * \brief Close a socket.
 *
 * wlp_close() will close and remove a socket that was created with
 * wlp_socket(). All memory used by the socket will be free'd.
 *
 * A connected socket will be cleanly disconnected from the remote end.
 * If there is still pending data on the socket (wlp_recv() was not called to
 * fetch all data), a TCP RST will be sent to the remote end.
 *
 * A socket can be closed form any state. Any connection callbacks registed for
 * the socket (e.g. from wlp_connect(), wlp_accept() or wlp_listen()) will not
 * be invoked when the socket is closed.
 * 
 * @param sockid of the socket to close.
 * @return
 *   - 0 on success
 *
 */
int wlp_close(int sockid);


/**
 * \brief Read data from a socket.
 * 
 * Read pending data from a socket. This function will attempt to
 * read up to len bytes from the given socket.
 *
 * If no data is pending on the socket, 0 will be returned.
 *
 * Sockets of type WLP_SOCK_STREAM will not preserve message boundaries in
 * general. If the socket has less than len bytes available for reading, the
 * available bytes will be read and the function will return. The actual number
 * of bytes read will be returned.
 *
 * Sockets of type WLP_SOCK_DGRAM and WLP_SOCK_RAW will preserve
 * message boundaries by requiring that the len parameter is large enough
 * to hold the complete message. Otherwise WLP_ERR_SIZE will be returned and
 * no data will be received. In this case, the caller must invoke wlp_recv()
 * again with a larger buffer that can hold the complete message. Also note that
 * any incoming packets addressed to a socket of type WLP_SOCK_DGRAM or
 * WLP_SOCK_RAW that are larger than 1500 bytes (including IP/UDP header) will
 * be discarded by the device.
 *
 * In case the host wants to discard a packet but do not have enough memory to
 * hold the complete packet, a NULL pointer can be passed as the buf parameter.
 * This pointer will be passed along to the bus read and write functions
 * registered in wlp_init(). Therefore, this assumes that these user-provided
 * functions can handle the NULL pointer and that the semantics is to read
 * data (according to the len paramter) but discard the bytes instead of writing
 * the destination buffer.
 *
 * For sockets of type WLP_SOCK_RAW, the IP header will be included in the
 * received message.
 *
 * See wlp_connect() for more information about message boundaries for the
 * differenct socket types.
 *
 * This function should typically be called from the recv_cb registered through
 * wlp_set_recv_cb(). However it is also valid to wlp_recv() even through a
 * recv_cb is not invoked for the socket (or not registerd at all). Therefore,
 * it is possible to recieve data from a specific socket in a polling fashion.
 *
 * Note that the error code WLP_ERR_CONN will be returned if the socket was
 * closed from the remote end. Note that any pending, buffered, data will be
 * returned before WLP_ERR_CONN is returned. In case the socket connection was
 * closed by the remote end, any registered recv_cb (see wlp_set_recv_cb())
 * will also be called with len set to 0.
 *
 * The error code WLP_ERR_ABORT will be returned if the socket connection
 * was terminated in any other way, e.g. link goes down. In this case any
 * registered conn_cb (see wlp_connect() or wlp_accept()) will also be called
 * with the connected parameter set to 0. Any non-read data that is buffered
 * by the wifi device will be lost.
 * 
 * @param sockid is the socket to read
 * @param buf should point to a buffer which will hold the read data.
 * @param len is the number of bytes to read.
 * @return
 *   - The number of bytes read on success
 *   - WLP_ERR_PARAM if invalid parameters, e.g sockid not found
 *   - WLP_ERR_MEM if out of memory or resources on the wifi device
 *   - WLP_ERR_CONN if the socket was disconnected by the remote end and all
 *              buffered data has been read or if the socket has not been
 *              connected.
 *   - WLP_ERR_ABORT if the socket connection was lost.
 *   - WLP_ERR_STATE if the socket is in an invalid state (e.g. listening).
 *   - WLP_ERR_SIZE if the complete message does not fit into the user
 *              supplied buffer. Only for sockets of type WLP_SOCK_DGRAM or
 *              WLP_SOCK_RAW.
 *   - WLP_ERR_PROTO if unrecoveraly protocol error. If this happens, there
 *              is probably a problem in the serial commuincation between the
 *              host and the wifi device.
 *
 */
int wlp_recv(int sockid, char *buf, int16_t len);


/**
 * \brief Send data to a socket.
 *
 * Send data to a connected socket.
 *
 * Sockets of type WLP_SOCK_STREAM will not preserve message boundaries in
 * general.  This function will attempt to send up to len bytes of data to the
 * given socket. If the socket has room for less than
 * len bytes in its input buffer, the maximum allowed bytes will be written
 * and then this function will return. The actual number of bytes written will
 * be returned.
 *
 * Sockets of type WLP_SOCK_DGRAM and WLP_SOCK_RAW will preserve
 * message boundaries by ensuring that the requested len is also the length of
 * the single IP packet transmitted over the wireless link (possibly fragmented)
 * If the call to wlp_send() returns WLP_ERR_SIZE, the provide buffer is too
 * large to be transferred as a single message. No data is transferred in this
 * case. This limit is 1500 bytes.
 *
 * Note that for sockets of type WLP_SOCK_RAW, you can not modify the IP
 * headers (this is inconsistent with the data read through wlp_recv() where
 * you actually get the IP headers), you can only specify the IP payload here.
 *
 * See wlp_connect() for more information about message boundaries for the
 * differenct socket types.
 *
 * It is valid to call wlp_send() multple times even though all requested data
 * was not transmitted (according to the return value) to send a large chunk
 * of data. Therefore, it is possible to send data to specific socket in a
 * blocking fashion.
 *
 * Note that the error code WLP_ERR_CONN will be returned if the socket was
 * closed by the remote end. In case the socket connection was closed
 * by the remote end, any registered recv_cb (see wlp_set_recv_cb())
 * will also be called with len set to 0.
 *
 * The error code WLP_ERR_ABORT will be returned if the socket connection
 * was terminated in any other way, e.g. link goes down. In this case the
 * conn_cb will also be called with the connected parameter set to 0.
 *
 * @param sockid is the socket to write
 * @param buf should pointer to a buffer which will hold the data to be written.
 * @param len is the number of bytes to write.
 * @return
 *   - The number of bytes actually written on success
 *   - WLP_ERR_PARAM if invalid parameters, e.g sockid not found
 *   - WLP_ERR_MEM if out of memory or resources on the wifi device
 *   - WLP_ERR_CONN if the socket was disconnected by the remote end (only
 *              for sockets of type WLP_SOCK_STREAM) or if the socket has not
 *              been connected.
 *   - WLP_ERR_ABORT if the socket connection was lost (only for sockets
 *              of type WLP_SOCK_STREAM).
 *   - WLP_ERR_STATE if the socket is in an invalid state (e.g. listening).
 *   - WLP_ERR_SIZE if the message is too long to pass atomically through
 *              the underlying protocol (IP). Only for sockets of type
 *              WLP_SOCK_DGRAM or WLP_SOCK_RAW.
 *   - WLP_ERR_PROTO if unrecoveraly protocol error. If this happens, there
 *              is probably a problem in the serial commuincation between the
 *              host and the wifi device.
 *   - WLP_FAILURE on any other error
 *
 */
int wlp_send(int sockid, const char *buf, int16_t len);


/**
 * \brief Send data to a specified address from a socket
 *
 * wlp_sendto() will send data to a specified address. This is generally used
 * for connectionless sockets, e.g. of type WLP_SOCK_DGRAM or WLP_SOCK_RAW.
 * For WLP_SOCK_TYPE_STREAM sockets, the IP and port parameters will be ignored
 * and the data will be sent to any remote endpoint established through
 * wlp_connect().
 *
 * For sockets of type WLP_SOCK_RAW, the port parameter will be ignored since
 * it is not included in the IP header.
 *
 * Except for the extra parameters, this function will behave exactly the same
 * as wlp_send(). See wlp_send() for more details.
 *
 * @param sockid is the socket to write
 * @param buf should pointer to a buffer which will hold the data to be written.
 * @param len is the number of bytes to write.
 * @param ip is the IP address to send to
 * @param port is the port to send to, ignored for sockets of type WLP_SOCK_RAW.
 * @return
 *   - The number of bytes actually written on success
 *   - WLP_ERR_PARAM if invalid parameters, e.g sockid not found
 *   - WLP_ERR_MEM if out of memory or resources on the wifi device
 *   - WLP_ERR_CONN if the socket was disconnected by the remote end (only
 *              for sockets of type WLP_SOCK_STREAM) or if the socket has not
 *              been connected.
 *   - WLP_ERR_ABORT if the socket connection was lost (only for sockets
 *              of type WLP_SOCK_STREAM).
 *   - WLP_ERR_STATE if the socket is in an invalid state (e.g. listening).
 *   - WLP_ERR_PROTO if unrecoveraly protocol error. If this happens, there
 *              is probably a problem in the serial commuincation between the
 *              host and the wifi device.
 *   - WLP_ERR_SIZE if the message is too long to pass atomically through
 *              the underlying protocol (IP). Only for sockets of type
 *              WLP_SOCK_DGRAM and WLP_SOCK_RAW.
 *   - WLP_FAILURE on any other error
 *
 */
int wlp_sendto(int sockid, const char *buf, int16_t len,
               const struct ip_addr *ip, uint16_t port);


/**
 * \brief Get name of connected peer socket.
 *
 * wlp_get_peeraddr() returns the address of the peer connected to the socket.
 * Only supported on sockets of type WLP_SOCK_STREAM.
 *
 * @param sockid is the socket for which the peer should be obtained.
 * @param peer will hold the peer IP address upon successful return.
 * @return
 *   - 0 on success
 *   - WLP_ERR_PARAM if invalid parameters, e.g sockid not found
 *   - WLP_ERR_CONN if the socket was disconnected by the remote end or if
 *              the socket has not been connected.
 *   - WLP_ERR_ABORT if the socket connection was lost (only for sockets
 *              of type WLP_SOCK_STREAM).
 *   - WLP_ERR_STATE if the socket is in an invalid state (e.g. listening).
 *   - WLP_ERR_NOTSUPP if the operation is not supported on the socket
 *
 */
int wlp_get_peeraddr(int sockid, struct ip_addr *peer);

/**
 * \brief Get network address given a dns name
 * 
 * If name is an IPv4 address, no lookup is performed and wlp_get_hostbyname()
 * copies host into the parameter provided to the lookup_cb.
 *
 * Every time wlp_poll() is invoked, the callback condition will be evaluated
 * and any registered callback will be invoked from same context as the call to
 * wlp_poll().
 *
 * @param host is the dns name to lookup
 * @param lookup_cb will be invoked when lookup is complete or considered
 *              failed. On success, the IP parameter will be set to the IP
 *              address that was looked up. On failure, IP will be NULL.
 * @param ctx is an opaque context parameter that will be passed back to the
 *        lookup_cb.
 *
 * @return
 *   - 0 on success
 *   - WLP_ERR_PARAM if invalid parameters
 *   - WLP_ERR_MEM if out of resources on device
 *   - WLP_ERR_BUSY if a previous call to wlp_get_hostbyname() is not
 *              yet completed.
 */
int wlp_get_hostbyname(const char *host,
                       void (*lookup_cb)(void *ctx, const struct ip_addr *ip),
                       void *ctx);

#endif /* WLP_STACK_OFFLOAD */

/**
 * \brief Event progress function.
 *
 * This function must be called in stand-alone environments to
 * ensure forward progress of events. The call can be made periodically from
 * the main application loop.
 *
 * Every time wlp_poll() is invoked, the callback conditions for all callbacks
 * registered through wlp_api will be evaluated and any registered callback
 * will be invoked from same context as the call to wlp_poll().
 *
 * If no callbacks are registered, wlp_poll() does not have to be invoked.
 *
 * In systems using an OS, a task could be scheduled to run this task
 * periodicallly.
 *
 * Note that this function should not be invoked from interrupt context.
 *
 *
 */
void wlp_poll(void);

/*! @} */



/*! \defgroup wlp_ap Access point functions
 *
 * These functions manage the access point mode.
 *
 * @{
 *
 */ 


/**
 * \brief Set the Wi-Fi mode to AP or STA.
 *
 * wlp_set_mode() will configure the Wi-Fi device to operate either as a
 * station or as an access point. The current mode can be obtained through
 * wlp_get_mode(). Note that the mode configured through wlp_set_mode() only
 * takes effect after wlp_reset() has been called. The configured mode persists
 * after power-cycling the device.
 *
 * @param mode is the mode to configure. This must be either WL_MODE_STA or
 *              WL_MODE_AP.
 *
 * @return
 *   - 0 on success
 *   
 */
int wlp_set_mode(uint8_t mode);



/**
 * \brief Check if the device is currently in AP or STA mode.
 *
 * wlp_get_mode() will be return the current operation mode of the Wi-Fi device.
 * Since the operation mode is persistent between power-cycles, it is generally
 * a good idea to call this function at startup to ensure the desired operation
 * mode.
 *
 * @param mode will hold the current mode upon successful return. This will be
 *              either WL_MODE_STA or WL_MODE_AP.
 *
 * @return
 *   - 0 on success
 *   
 */
int wlp_get_mode(uint8_t *mode);



/**
 * \brief Set the channel. Only supported when mode is WLP_MODE_AP.
 *
 * This function is only supported when wlp_get_mode() returns WLP_MODE_AP.
 * wlp_set_channel() must be called before wlp_linkup() to have effect.
 *
 * @return
 *   - 0 on success
 *   - WLP_ERR_PARAM if invalid channel
 *   - WLP_ERR_STATE the operation can not be carried out (e.g. ap already
 *              started).
 *   - WLP_ERR_NOTSUPP if operation is not supported
 */
int wlp_set_channel(uint8_t channel);


/**
 * \brief Configure parameters for access point mode
 *
 * Configure the parameters that will be used when the device is in access
 * point mode, WL_MODE_AP.
 *
 * @param hidden can be set to non-zero in order to configure the device
 *               with hidden SSID. This means that the device will not
 *               broadcast its SSID, but it will still respond to clients
 *               that already knows the SSID.
 * @param beacon_intvl specifies interval in which beacons are transmitted
 *               by the device. This parameter does not have any effect if
 *               hidden is 0
 * @param inact_tmo specifies the idle time in minutes before the device
 *               will disconnect an inactive client.
 * @param country specifies the region that should be advertised by the device.
 *               Will not be changed if NULL.
 *
 * @return
 *   - 0 on success
 *   - A negative error code on failure
 *
 */
int
wlp_conf_ap(int hidden, int beacon_intvl, int inact_tmo, const char *country);

/*! @} */

/*! \defgroup wlp_ps Functions to manage power consumption
 *
 * These functions configure features that affect power consumption.
 * Achieving low power operation is always a tradeoff between some
 * aspect of performance (or an entire feature) and the power
 * consumption of the device. Beware that these functions can make the
 * device difficult to reach. It is recommended that you verify the
 * functionality of your application before attempting to reduce the
 * power consumption with these functions.
 *

In order to reduce the power consumption of the whole system
there are three different domains to consider.

\li \ref radios
\li \ref wifi_device_ps
\li \ref host_ps

\section radios WiFi Radio Power Save

The WiFi radio power save modes are controlled by \a wlp_conf_ps()
and \a wlp_set_ps().  The WiFi will, if so commanded by \a
wlp_set_ps(), sleep whenever possible and will only wake up and turn
on its receiver and transmitter when necessary.

\b Performance \b impact : Turning on WiFi power save will increase the
average receive latency since the WiFi only gets notifications of
pending packets at most once every beacon interval (which is usually
100ms).

\b Example : Enabling the WiFi power save mode

\code
if (wlp_conf_ps(0,    // use_ps_poll 
                200,  // ps_traffic_timeout 
                2000, // ps_delay 
                1,    // rx_all_dtim 
                1     // listen_interval
            ) < 0) {
        handle_error();
        return;
}
// Enable WiFi power save
wlp_set_ps(1);
\endcode

\section wifi_device_ps WiFi Device Sleep


<h2>Using the WAKEUP pin</h2>

The WiFi device can only sleep if it can be woken up by the host when
the host wants to send a command or poll events. A requirement for the
WiFi device to enter sleep mode is the presence of the WAKEUP pin
which is used to wake the device whenever the host wants to
communicate with it. The WAKEUP pin is not necessary for correct
operation but the presence of the pin will cause the device to consume
significantly less power. 

\b Performance \b impact : A slight delay is introduced in the communication
with the device.

\b Pin \b usage

 1. Connect the WAKEUP pin on the WiFi device to a pin on the host
    that the application can control. The WAKEUP pin is an output pin
    on the host, active low, and has a pull-up resistor on the WiFi
    device.

 2. Write a function wakeup_cb(void *ctx, int enable) that sets 
    the WAKEUP pin if enable is 0 and clears the WAKEUP pin if enable is 1.
    The pin is active low which is why the logic may seem inverted.

 3. Register the wakeup_cb() function with the wlp_api library using
    \a wlp_set_wakeup_cb(). The library will then automatically use the
    wakeup_cb() function to wake up the WiFi device whenever necessary.

\b Example : Supporting the WAKEUP pin

\code
static void wakeup_cb(void *ctx, int enable) {
        if (enable) {
                // Signal wakeup (active low)
                owlgpio_clr(OWL_BOARD_SPB800P_WAKEUP_PIN);
                // Wait for the device to wake up
                owlboard_mdelay(30);
        }
        else {
                // Reset wakeup (active low)
                owlgpio_set(OWL_BOARD_SPB800P_WAKEUP_PIN);
        }
}

// Register a WAKEUP pin control callback.
wlp_set_wakeup_cb(wakeup_cb, sh);
\endcode

<h2>Disabling LEDs</h2>

If the WiFi device has LEDs those can be disabled to conserve power.

\b Performance \b impact : None.

\b Example : Disable LEDs

\code
wlp_set_led(0);
\endcode

<h2>Reducing the WiFi Device Internal Polling Interval</h2>

The IP stack on the WiFi device must be polled periodically. For TCP
compliance the stack must be polled at least once every 250ms. In
situations where minimum power consumption is desired and TCP is not
used the polling interval can be increased.

\b Performance \b impact : DHCP and ARP timeouts are adversely affected. TCP
functionality may be impaired. The poll period should not be increased
beyond 250ms unless the traffic patterns are well known and known to
work with the decreased timer resolution.

\b Example : Set the polling interval to 5s

\code
wlp_set_poll_period(5000);
\endcode

\section host_ps Host Sleep

The host must normally poll the WiFi device to get notifications of
asynchronous events (such as traffic pending from the WiFi).  This
requires the host to wake up periodically for the polling
operation. To allow the host to sleep longer a HOST_ATTENTION pin is
supported on some hardware revisions. If the WiFi device supports the
HOST_ATTENTION pin it will signal on this pin whenever events are
pending. The host only has to poll the WiFi device when the
HOST_ATTENTION pin is signalled. This allows the host to sleep
whenever the pin is not signalled.

\b Performance \b impact : None.

<h2>Using the HOST_ATTENTION pin</h2>

 1. Connect the HOST_ATTENTION pin on the WiFi device to a pin on the host that
    the application can control. The HOST_ATTENTION pin is an input pin
    on the host, active low, and should have a pull-up resistor on the
    host.

 2. Check if the WiFi device supports the HOST_ATTENTION pin by
    calling wlp_get_fw_caps(&caps) and checking if the flag WLP_CAP_HA_PIN
    is set in caps.

 3. Make the application check the status of the HOST_ATTENTION pin as
    frequently as possible. Whenever the HOST_ATTENTION pin is low
    then \a wlp_poll() should be called until the HOST_ATTENTION pin goes
    high.  This is different from the case when the HOST_ATTENTION pin
    is unavailable/unused (such as in legacy code, or when the
    WLP_CAP_HA_PIN is not set in the fw capabilities flag field), when
    the \a wlp_poll() function should be called unconditionally.

\b Example : Supporting the HOST_ATTENTION pin

\code
static int ha_pin_is_signalled(void)
{
        if (owlgpio_is_low(OWL_BOARD_SPB800P_HOST_ATTENTION_PIN))
                return 1;
        return 0;
}

uint8_t fw_caps;
wlp_get_fw_caps(&fw_caps);

for (;;) {
        ...
        if (fw_caps & WLP_CAP_HA_PIN) {
                // If the host attention pin is present we
                // only poll if the pin is asserted (active low).
                // When the pin is asserted there's an event
                // pending from the SPB800 module.
                if (ha_pin_is_signalled())
                        wlp_poll();
        }
        else {
                // Without the host attention pin we need to
                // poll the SPB800 periodically. This method will
                // prevent the SPB800 from using power save
                // level 1.
                wlp_poll();
        }
        ...
}                
\endcode

 * @{
 *
 */ 

/**
 * \brief Register a callback that will be called to wakeup the wifi device.
 *
 * The callback will be invoked to control a wakeup-pin if one is
 * present on the current host platform. The wakeup-pin, if present,
 * is connected to the wifi device and is required if the wifi device
 * should be able to operate in the lowest power mode (power save
 * level 1). If the \a wakeupf() function is called with enable set to
 * 1 the wakeup-pin must be pulled low and if enable is set to 0 the
 * wakeup-pin must be pulled high. If the \a wakeupf() callback is
 * not provided then the wifi device can only operate in power save
 * level 2.
 *
 * @param wakeupf Wakeup-pin control function.
 * @param ctx is an opaque context pointer that will be passed back to readf,
 *              writef and selectf.
 * 
 * @return
 *   - 0 on success
 *
 */
int wlp_set_wakeup_cb(void (*wakeupf)(void *ctx, int enable),
                      void *ctx);

/** \brief Enable or disable WiFi power save.
 * 
 * Enable/disable legacy power save mode (as opposed to WMM power save
 * which is not supported in the current driver). In legacy power save
 * mode, the device will power down when idle. When connected, the
 * device will wake up to receive beacon frames and any buffered data
 * from the AP. The response time when legacy power save is enabled
 * might therefore be as long as the AP beacon interval (commonly 100
 * ms). However, the throughput should not be affected.
 *
 * @param enable determines if WiFi power save should be enabled or disabled.
 * @return
 *    - 0 on success
 */
int wlp_set_ps(int enable);

/** \brief Configure WiFi power save parameters.
 * 
 * This call is not necessary to enable WiFi power save.
 * If \a wlp_set_ps(1) is called before calling \a wlp_conf_ps()
 * then default power save parameters are used.
 *
 * Beware that setting \a rx_all_dtim to 0 and \a listen_interval to a
 * large value (> 10) can make the device very difficult to
 * reach. When \a rx_all_dtim is 0 you risk loosing multicast data
 * because that data is transferred after a DTIM beacon. This will be
 * particularly noticable when another entitity on the same network
 * segment tries to reach the device and does not know which MAC
 * address the device has. In this case an ARP packet is normally sent
 * to the broadcast address and this will be unlikely to be heard
 * by the device.
 *
 * @param use_ps_poll Use PS-Poll frames to retrieve buffered data.
 *        Any changes to this parameter will take effect upon next
 *        connect or when power save is enabled through
 *        wlp_set_ps().
 *        Note: To retrieve one buffered packet, the ps poll scheme
 *        needs one ps poll packet to the AP instead of two null
 *        packets in the power management bit scheme. Ps poll avoids
 *        the overhead of traffic monitoring time in active mode as
 *        well. But since each ps poll request can make the AP release
 *        only one buffered packet, it is not the optimal scheme for
 *        applications with heavy downlink traffic.
 * @param ps_traffic_timeout Timeout in [ms] to wait for more buffered
 *        data from AP. This setting has no effect if use_ps_poll is
 *        1. Any changes to this parameter will take effect
 *        immediately.
 * @param ps_delay Power save will de delayed ps_delay [ms] after
 *        connecting to an AP.
 * @param rx_all_dtim If set to 1, then STA will wake up to listen to
 *        every beacon containing DTIM (delivery traffic indication
 *        messages) when connected. The actual DTIM interval is
 *        configured in the AP.  If the DTIM interval, as configured
 *        in the AP, is larger than \a listen_interval, the STA will
 *        wakeup according to the \a listen_interval parameter.  Note
 *        that if \a rx_all_dtim is set to 0 and \a listen_interval is
 *        set to something larger than 1 you risk missing reception of
 *        broadcast/multicast data because the AP will transmit that
 *        in conjunction with a DTIM beacon.
 * @param listen_interval The Listen Interval field is used to
 *        indicate to the AP how often a STA in power save mode wakes
 *        to listen to beacon frames. The value of this parameter is
 *        expressed in units of Beacon Interval. An AP may use the
 *        Listen Interval information in determining the lifetime of
 *        frames that it buffers for a STA.  Any changes to this
 *        parameter will take effect upon next association.
 *
 * @return 
 *       - 0 on success. 
 *       - WLP_FAILURE if the configuration failed (most likely
 *         due to invalid parameters).
 */
int wlp_conf_ps(uint8_t use_ps_poll,
                uint32_t ps_traffic_timeout,
                uint32_t ps_delay,
                uint8_t rx_all_dtim,
                uint16_t listen_interval);


/**
 * \brief Enable/disable LEDs on the device.
 *
 * The 2 LEDs that light when the WiFi link and the IP link is up can
 * be disabled to conserve power.
 *
 * This function does not control the state of the LEDs (it cannot be
 * used to light the LEDs). If called with \a enable set to 0 it will
 * disable the LEDs completely. If called with \a enable set to 1 the
 * LEDs will be controlled by the internal link control logic (which
 * is the default behaviour).
 *
 * @param enable 1 to turn on the LED link indicator function. 0 to 
 *               turn LEDs completely off.
 * @return 
 *   - 0 on success
 */
int wlp_set_led(int enable);

/** 
 * \brief Configure the link heartbeat.
 *
 * Some access points disconnects stations that have been idle (that
 * have not sent any data) for some period of time. They may
 * disconnect the station explicitly or they may silently drop the
 * connection.  This can cause problems if the device is a passive
 * listener (waiting for data that may never come because the
 * connection has been silently dropped). The link heartbeat can be
 * used to work around this limitation.
 *
 * If the heartbeat \a period is > 0 then a NULL data frame will be
 * sent to the access point at the specified \a period whenever a WiFi
 * connection has been established.
 *
 * Bear in mind that a low value for \a period will increase the power
 * consumption of the device. A reasonable value is 60000 or 120000
 * (sending heartbeats every 1 or 2 minutes).
 *
 * This function does not do anything in the AP-mode.
 *
 * The default heartbeat period is 60s.
 *
 * @param period The heartbeat period in milliseconds. 0 will disable
 *               the heartbeat functionality.
 * @return
 *   - 0 on success
 */
int wlp_set_heartbeat(uint32_t ms_period);

/**
 * \brief Configure the device IP-stack poll frequency.
 *
 * The device must wake up periodically to run timers required by
 * the IP stack. The wake up frequency affects power consumption.  To
 * provide a IETF-compliant TCP connection the poll period must be
 * 250ms or less. If no TCP connection is active, or if only UDP/RAW
 * data is handled, then it may be possible to increase the poll
 * period to a couple of seconds to reduce power consumption.
 *
 * Be aware that long polling periods can have large and
 * hard-to-predict consequences on network connectivity at the IP
 * layer and above. 
 *
 * This function only controls the polling frequency of the IP stack
 * in the device itself. It is unrelated to the \a wlp_poll() function
 * which is running on the host, not on the device.
 *
 * Note that when the wakeup-pin is supported (see \a
 * wlp_set_wakeup_cb()) and asserted then the device is kept awake (and
 * continuously checking pending timers) irrespective of the poll
 * period. The poll period is only used when the device actually
 * sleeps.
 *
 * @param ms_period Poll period in ms.
 * @return
 *   - 0 on success
 */

int wlp_set_poll_period(uint32_t ms_period);

/*! @} */

#endif /* WLP_API_H */
